# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""unit tests for module cubicweb.schema"""

from os.path import join, dirname, splitext
import json

from logilab.common.testlib import TestCase, unittest_main
from rql import RQLSyntaxError
from yams import ValidationError, BadSchemaDefinition
from yams.buildobjs import (
    RelationDefinition,
    EntityType,
    RelationType,
    Int,
    String,
    ComputedRelation,
)
from yams.reader import fill_schema

from cubicweb.devtools.apptest_config import ApptestConfiguration
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.schema import (
    CubicWebSchema,
    CubicWebSchemaLoader,
    RQLConstraint,
    RQLUniqueConstraint,
    RQLVocabularyConstraint,
    ERQLExpression,
    RRQLExpression,
    normalize_expression,
    order_eschemas,
    guess_rrqlexpr_mainvars,
    build_schema_from_namespace,
)
from cubicweb.schema_exporters import TypescriptSchemaExporter, JSONSchemaExporter

DATADIR = join(dirname(__file__), "data")

# build a dummy schema ########################################################


PERSONNE_PERMISSIONS = {
    "read": ("managers", "users", "guests"),
    "update": ("managers", "owners"),
    "add": ("managers", ERQLExpression("X travaille S, S owned_by U")),
    "delete": (
        "managers",
        "owners",
    ),
}

CONCERNE_PERMISSIONS = {
    "read": ("managers", "users", "guests"),
    "add": ("managers", RRQLExpression("U has_update_permission S")),
    "delete": ("managers", RRQLExpression("O owned_by U")),
}

schema = CubicWebSchema("Test Schema")
enote = schema.add_entity_type(EntityType("Note"))
eaffaire = schema.add_entity_type(EntityType("Affaire"))
eperson = schema.add_entity_type(
    EntityType(
        "Personne",
        description="A person with a name",
        __permissions__=PERSONNE_PERMISSIONS,
    )
)
esociete = schema.add_entity_type(EntityType("Societe"))

RELS = (
    # attribute relations
    ("Note date String"),
    ("Note type String"),
    ("Affaire sujet String"),
    ("Affaire ref String"),
    ("Personne nom String"),
    ("Personne prenom String"),
    ("Personne sexe String"),
    ("Personne tel Int"),
    ("Personne fax Int"),
    ("Personne datenaiss Date"),
    ("Personne promo String"),
    # real relations
    ("Personne  travaille Societe"),
    ("Personne  evaluee   Note"),
    ("Societe evaluee   Note"),
    ("Personne  concerne  Affaire"),
    ("Personne  concerne  Societe"),
    ("Affaire concerne  Societe"),
)
done = {}
for rel in RELS:
    _from, _type, _to = rel.split()
    if _type.lower() not in done:
        schema.add_relation_type(RelationType(_type))
        done[_type.lower()] = True
    if _type == "concerne":
        schema.add_relation_def(
            RelationDefinition(_from, _type, _to, __permissions__=CONCERNE_PERMISSIONS)
        )
    else:
        schema.add_relation_def(RelationDefinition(_from, _type, _to))

schema.add_relation_type(RelationType("colabore_avec", symmetric=True))
done["colabore_avec"] = True
schema.add_relation_def(RelationDefinition("Personne", "colabore_avec", "Societe"))


class CubicWebSchemaTC(TestCase):
    def test_rql_constraints_inheritance(self):
        # isinstance(cstr, RQLVocabularyConstraint)
        # -> expected to return RQLVocabularyConstraint and RQLConstraint
        #   instances but not RQLUniqueConstraint
        #
        # isinstance(cstr, RQLConstraint)
        # -> expected to return RQLConstraint instances but not
        #    RQLVocabularyConstraint and RQLUniqueConstraint
        self.assertFalse(issubclass(RQLUniqueConstraint, RQLVocabularyConstraint))
        self.assertFalse(issubclass(RQLUniqueConstraint, RQLConstraint))

    def test_entity_perms(self):
        self.assertEqual(eperson.get_groups("read"), {"managers", "users", "guests"})
        self.assertEqual(
            eperson.get_groups("update"),
            {
                "managers",
                "owners",
            },
        )
        self.assertEqual(eperson.get_groups("delete"), {"managers", "owners"})
        self.assertEqual(eperson.get_groups("add"), {"managers"})
        self.assertEqual(
            [str(e) for e in eperson.get_rqlexprs("add")],
            ["Any X WHERE X travaille S, S owned_by U, X eid %(x)s, U eid %(u)s"],
        )
        eperson.set_action_permissions("read", ("managers",))
        self.assertEqual(eperson.get_groups("read"), {"managers"})

    def test_relation_perms(self):
        rconcerne = schema.relation_schema_for("concerne").relation_definition(
            "Personne", "Societe"
        )
        self.assertEqual(rconcerne.get_groups("read"), {"managers", "users", "guests"})
        self.assertEqual(rconcerne.get_groups("delete"), {"managers"})
        self.assertEqual(rconcerne.get_groups("add"), {"managers"})
        rconcerne.set_action_permissions("read", ("managers",))
        self.assertEqual(rconcerne.get_groups("read"), {"managers"})
        self.assertEqual(
            [str(e) for e in rconcerne.get_rqlexprs("add")],
            ["Any S,U WHERE U has_update_permission S, S eid %(s)s, U eid %(u)s"],
        )

    def test_erqlexpression(self):
        self.assertRaises(RQLSyntaxError, ERQLExpression, "1")
        expr = ERQLExpression("X travaille S, S owned_by U")
        self.assertEqual(
            str(expr),
            "Any X WHERE X travaille S, S owned_by U, X eid %(x)s, U eid %(u)s",
        )
        expr = ERQLExpression("X foo S, S bar U, X baz XE, S quux SE HAVING XE > SE")
        self.assertEqual(
            str(expr),
            "Any X WHERE X foo S, S bar U, X baz XE, S quux SE, X eid %(x)s, "
            "U eid %(u)s HAVING XE > SE",
        )

    def test_rrqlexpression(self):
        self.assertRaises(Exception, RRQLExpression, "1")
        self.assertRaises(RQLSyntaxError, RRQLExpression, "O X Y")
        expr = RRQLExpression("U has_update_permission O")
        self.assertEqual(
            str(expr),
            "Any O,U WHERE U has_update_permission O, O eid %(o)s, U eid %(u)s",
        )


loader = CubicWebSchemaLoader()
config = ApptestConfiguration("data", __file__)
config.bootstrap_cubes()


class SchemaReaderClassTest(TestCase):
    def test_order_eschemas(self):
        schema = loader.load(config)
        self.assertEqual(
            order_eschemas([schema["Note"], schema["SubNote"]]),
            [schema["Note"], schema["SubNote"]],
        )
        self.assertEqual(
            order_eschemas([schema["SubNote"], schema["Note"]]),
            [schema["Note"], schema["SubNote"]],
        )

    def test_knownValues_load_schema(self):
        schema = loader.load(config)
        self.assertIsInstance(schema, CubicWebSchema)
        self.assertEqual(schema.name, "data")
        entities = sorted([str(e) for e in schema.entities()])
        expected_entities = [
            "Ami",
            "BaseTransition",
            "BigInt",
            "Bookmark",
            "Boolean",
            "Bytes",
            "Card",
            "Comment",
            "Date",
            "Datetime",
            "Decimal",
            "CWComputedRType",
            "CWConstraint",
            "CWConstraintType",
            "CWDataImport",
            "CWEType",
            "CWAttribute",
            "CWGroup",
            "EmailAddress",
            "CWRelation",
            "CWPermission",
            "CWProperty",
            "CWRType",
            "CWSession",
            "CWSource",
            "CWSourceHostConfig",
            "CWUniqueTogetherConstraint",
            "CWUser",
            "ExternalUri",
            "FakeFile",
            "Float",
            "Int",
            "Interval",
            "Note",
            "Password",
            "Personne",
            "Produit",
            "RQLExpression",
            "Reference",
            "Service",
            "Societe",
            "State",
            "StateFull",
            "String",
            "SubNote",
            "SubWorkflowExitPoint",
            "Tag",
            "TZDatetime",
            "TZTime",
            "Time",
            "Transition",
            "TrInfo",
            "Usine",
            "Ville",
            "Workflow",
            "WorkflowTransition",
        ]
        self.assertListEqual(sorted(expected_entities), entities)
        relations = sorted([str(r) for r in schema.relations()])
        expected_relations = [
            "actionnaire",
            "add_permission",
            "address",
            "alias",
            "allowed_transition",
            "associe",
            "bookmarked_by",
            "by_transition",
            "buddies",
            "cardinality",
            "comment",
            "comment_format",
            "comments",
            "composite",
            "condition",
            "config",
            "connait",
            "constrained_by",
            "constraint_of",
            "content",
            "content_format",
            "contrat_exclusif",
            "created_by",
            "creation_date",
            "cstrtype",
            "custom_workflow",
            "cwuri",
            "cwsessiondata",
            "cw_import_of",
            "cw_host_config_of",
            "cw_source",
            "data",
            "data_encoding",
            "data_format",
            "data_name",
            "default_workflow",
            "defaultval",
            "delete_permission",
            "description",
            "description_format",
            "destination_state",
            "dirige",
            "ean",
            "ecrit_par",
            "eid",
            "end_timestamp",
            "evaluee",
            "expression",
            "exprtype",
            "extra_props",
            "fabrique_par",
            "final",
            "firstname",
            "for_user",
            "formula",
            "fournit",
            "from_entity",
            "from_state",
            "fulltext_container",
            "fulltextindexed",
            "has_group_permission",
            "has_text",
            "identity",
            "in_group",
            "in_state",
            "in_synchronization",
            "indexed",
            "initial_state",
            "inlined",
            "insee",
            "internationalizable",
            "is",
            "is_instance_of",
            "label",
            "last_login_time",
            "latest_retrieval",
            "lieu",
            "log",
            "login",
            "mainvars",
            "match_host",
            "modification_date",
            "name",
            "nom",
            "ordernum",
            "owned_by",
            "parser",
            "path",
            "pkey",
            "prefered_form",
            "prenom",
            "promo",
            "primary_email",
            "read_permission",
            "relation_type",
            "relations",
            "require_group",
            "rule",
            "specializes",
            "start_timestamp",
            "state_of",
            "status",
            "subworkflow",
            "subworkflow_exit",
            "subworkflow_state",
            "surname",
            "symmetric",
            "synopsis",
            "tags",
            "title",
            "to_entity",
            "to_state",
            "transition_of",
            "travaille",
            "type",
            "upassword",
            "update_permission",
            "url",
            "uri",
            "use_email",
            "value",
            "wf_info_for",
            "wikiid",
            "workflow_of",
            "tr_count",
        ]

        self.assertListEqual(sorted(expected_relations), relations)

        eschema = schema.entity_schema_for("CWUser")
        rels = sorted(str(r) for r in list(eschema.subject_relations.values()))
        self.assertListEqual(
            rels,
            [
                "buddies",
                "created_by",
                "creation_date",
                "custom_workflow",
                "cw_source",
                "cwuri",
                "eid",
                "evaluee",
                "firstname",
                "has_group_permission",
                "has_text",
                "identity",
                "in_group",
                "in_state",
                "is",
                "is_instance_of",
                "last_login_time",
                "login",
                "modification_date",
                "owned_by",
                "primary_email",
                "surname",
                "upassword",
                "use_email",
            ],
        )
        rels = sorted(r.type for r in list(eschema.object_relations.values()))
        self.assertListEqual(
            rels,
            [
                "bookmarked_by",
                "buddies",
                "created_by",
                "for_user",
                "identity",
                "owned_by",
                "wf_info_for",
            ],
        )
        rschema = schema.relation_schema_for("relation_type")
        properties = rschema.relation_definition("CWAttribute", "CWRType")
        self.assertEqual(properties.cardinality, "1*")
        constraints = properties.constraints
        self.assertEqual(len(constraints), 1, constraints)
        constraint = constraints[0]
        self.assertIsInstance(constraint, RQLConstraint)
        self.assertEqual(constraint.expression, "O final TRUE")

    def test_fulltext_container(self):
        schema = loader.load(config)
        self.assertIn("has_text", list(schema["CWUser"].subject_relations.values()))
        self.assertNotIn(
            "has_text", list(schema["EmailAddress"].subject_relations.values())
        )

    def test_permission_settings(self):
        schema = loader.load(config)
        aschema = schema["TrInfo"].relation_definition("comment")
        self.assertEqual(aschema.get_groups("read"), {"managers", "users", "guests"})
        self.assertEqual(aschema.get_rqlexprs("read"), ())
        self.assertEqual(aschema.get_groups("update"), {"managers"})
        self.assertEqual(
            [x.expression for x in aschema.get_rqlexprs("update")],
            ["U has_update_permission X"],
        )

    def test_nonregr_allowed_type_names(self):
        schema = CubicWebSchema("Test Schema")
        schema.add_entity_type(EntityType("NaN"))

    def test_relation_perm_overriding(self):
        loader = CubicWebSchemaLoader()
        config = ApptestConfiguration("data_schemareader", __file__)
        config.bootstrap_cubes()
        schema = loader.load(config)
        rdef = next(iter(schema["in_group"].relation_definitions.values()))
        self.assertEqual(
            rdef.permissions,
            {"read": ("managers",), "add": ("managers",), "delete": ("managers",)},
        )
        rdef = next(iter(schema["cw_for_source"].relation_definitions.values()))
        self.assertEqual(
            rdef.permissions,
            {
                "read": ("managers", "users"),
                "add": ("managers",),
                "delete": ("managers",),
            },
        )

    def test_computed_attribute(self):
        """Check schema finalization for computed attributes."""

        class Person(EntityType):
            salary = Int()

        class works_for(RelationDefinition):
            subject = "Person"
            object = "Company"
            cardinality = "?*"

        class Company(EntityType):
            total_salary = Int(
                formula="Any SUM(SA) GROUPBY X WHERE P works_for X, P salary SA"
            )

        good_schema = build_schema_from_namespace(vars().items())
        rdef = good_schema["Company"].relation_definition("total_salary")
        # ensure 'X is Company' is added to the rqlst to avoid ambiguities, see #4901163
        self.assertEqual(
            str(rdef.formula_select),
            "Any SUM(SA) GROUPBY X WHERE P works_for X, P salary SA, X is Company",
        )
        # check relation definition permissions
        self.assertEqual(
            rdef.permissions,
            {"add": (), "update": (), "read": ("managers", "users", "guests")},
        )

        class Company(EntityType):  # noqa
            total_salary = String(
                formula="Any SUM(SA) GROUPBY X WHERE " "P works_for X, P salary SA"
            )

        with self.assertRaises(BadSchemaDefinition) as exc:
            build_schema_from_namespace(vars().items())

        self.assertEqual(
            str(exc.exception),
            "computed attribute total_salary on Company: "
            "computed attribute type (Int) mismatch with "
            "specified type (String)",
        )


class SchemaReaderComputedRelationAndAttributesTest(TestCase):
    def test_infer_computed_relation(self):
        class Person(EntityType):
            name = String()

        class Company(EntityType):
            name = String()

        class Service(EntityType):
            name = String()

        class works_for(RelationDefinition):
            subject = "Person"
            object = "Company"

        class produce(RelationDefinition):
            subject = ("Person", "Company")
            object = "Service"

        class achete(RelationDefinition):
            subject = "Person"
            object = "Service"

        class produces_and_buys(ComputedRelation):
            rule = "S produce O, S achete O"

        class produces_and_buys2(ComputedRelation):
            rule = "S works_for SO, SO produce O"

        class reproduce(ComputedRelation):
            rule = "S produce O"

        schema = build_schema_from_namespace(vars().items())

        # check object/subject type
        self.assertEqual(
            [("Person", "Service")],
            sorted(schema["produces_and_buys"].relation_definitions),
        )
        self.assertEqual(
            [("Person", "Service")],
            sorted(schema["produces_and_buys2"].relation_definitions),
        )
        self.assertCountEqual(
            [("Company", "Service"), ("Person", "Service")],
            sorted(schema["reproduce"].relation_definitions),
        )
        # check relation definitions are marked infered
        rdef = schema["produces_and_buys"].relation_definitions[("Person", "Service")]
        self.assertTrue(rdef.infered)
        # and have no add/delete permissions
        self.assertEqual(
            rdef.permissions,
            {"add": (), "delete": (), "read": ("managers", "users", "guests")},
        )

        class autoname(ComputedRelation):
            rule = "S produce X, X name O"

        with self.assertRaises(BadSchemaDefinition) as cm:
            build_schema_from_namespace(vars().items())
        self.assertEqual(str(cm.exception), "computed relations cannot be final")


class BadSchemaTC(TestCase):
    def setUp(self):
        self.loader = CubicWebSchemaLoader()
        self.loader.defined = {}
        self.loader.loaded_files = []
        self.loader.post_build_callbacks = []

    def _test(self, schemafile, msg):
        self.loader.handle_file(join(DATADIR, schemafile), splitext(schemafile)[0])
        sch = self.loader.schemacls("toto")
        with self.assertRaises(BadSchemaDefinition) as cm:
            fill_schema(sch, self.loader.defined, False)
        self.assertEqual(str(cm.exception), msg)

    def test_lowered_etype(self):
        self._test(
            "lowered_etype.py",
            "'my_etype' is not a valid name for an entity type. It should "
            "start with an upper cased letter and be followed by at least "
            "a lower cased letter",
        )

    def test_uppered_rtype(self):
        self._test(
            "uppered_rtype.py",
            "'ARelation' is not a valid name for a relation type. It should be lower cased",
        )

    def test_rrqlexpr_on_etype(self):
        self._test(
            "rrqlexpr_on_eetype.py",
            "can't use RRQLExpression on ToTo, use an ERQLExpression",
        )

    def test_erqlexpr_on_rtype(self):
        self._test(
            "erqlexpr_on_ertype.py",
            "can't use ERQLExpression on relation ToTo toto TuTu, use a RRQLExpression",
        )

    def test_rqlexpr_on_rtype_read(self):
        self._test(
            "rqlexpr_on_ertype_read.py",
            "can't use rql expression for read permission of relation ToTo toto TuTu",
        )

    def test_rrqlexpr_on_attr(self):
        self._test(
            "rrqlexpr_on_attr.py",
            "can't use RRQLExpression on attribute ToTo.attr[String], use an ERQLExpression",
        )

    def test_rqlexpr_on_computedrel(self):
        self._test(
            "rqlexpr_on_computedrel.py",
            "can't use rql expression for read permission of relation Subject computed Object",
        )


class NormalizeExpressionTC(TestCase):
    def test(self):
        self.assertEqual(
            normalize_expression("X  bla Y,Y blur Z  ,  Z zigoulou   X "),
            "X bla Y, Y blur Z, Z zigoulou X",
        )
        self.assertEqual(
            normalize_expression('X bla Y, Y name "x,y"'), 'X bla Y, Y name "x,y"'
        )


class RQLExpressionTC(TestCase):
    def test_comparison(self):
        self.assertEqual(
            ERQLExpression("X is CWUser", "X", 0), ERQLExpression("X is CWUser", "X", 0)
        )
        self.assertNotEqual(
            ERQLExpression("X is CWUser", "X", 0),
            ERQLExpression("X is CWGroup", "X", 0),
        )

    def test_has_update_permission(self):
        expr = ERQLExpression("P use_email X, U has_update_permission P")
        rql, found, keyarg = expr.transform_has_permission()
        self.assertEqual(rql, "Any X,P WHERE P use_email X, X eid %(x)s")
        self.assertEqual(found, [("update", 1)])
        self.assertEqual(keyarg, None)

    def test_expression(self):
        expr = ERQLExpression("EXISTS(U use_email X)")
        rql, found, keyarg = expr.transform_has_permission()
        self.assertEqual(
            rql, "Any X WHERE EXISTS(U use_email X), X eid %(x)s, U eid %(u)s"
        )
        self.assertEqual(found, None)
        self.assertEqual(keyarg, None)


class GuessRrqlExprMainVarsTC(TestCase):
    def test_exists(self):
        mainvars = guess_rrqlexpr_mainvars(
            normalize_expression(
                "NOT EXISTS(O team_competition C, C level < 3, C concerns S)"
            )
        )
        self.assertEqual(mainvars, {"S", "O"})


class RQLConstraintTC(CubicWebTC):
    def test_user_constraint(self):
        cstr = RQLConstraint("U identity O")
        with self.admin_access.repo_cnx() as cnx:
            anoneid = cnx.execute('Any X WHERE X login "anon"')[0][0]
            self.assertRaises(ValidationError, cstr.repo_check, cnx, 1, "rel", anoneid)
            self.assertEqual(
                cstr.repo_check(cnx, 1, cnx.user.eid), None
            )  # no validation error, constraint checked


class WorkflowShemaTC(CubicWebTC):
    def test_trinfo_default_format(self):
        with self.admin_access.cnx() as cnx:
            tr = cnx.user.cw_adapt_to("IWorkflowable").fire_transition("deactivate")
            self.assertEqual(tr.comment_format, "text/plain")


class CompositeSchemaTC(CubicWebTC):
    composites = {
        "Ami": [("comments", "Comment", "Ami", "object")],
        "BaseTransition": [("condition", "BaseTransition", "RQLExpression", "subject")],
        "CWAttribute": [
            ("add_permission", "CWAttribute", "RQLExpression", "subject"),
            ("constrained_by", "CWAttribute", "CWConstraint", "subject"),
            ("read_permission", "CWAttribute", "RQLExpression", "subject"),
            ("update_permission", "CWAttribute", "RQLExpression", "subject"),
        ],
        "CWEType": [
            ("add_permission", "CWEType", "RQLExpression", "subject"),
            ("constraint_of", "CWUniqueTogetherConstraint", "CWEType", "object"),
            ("delete_permission", "CWEType", "RQLExpression", "subject"),
            ("from_entity", "CWAttribute", "CWEType", "object"),
            ("from_entity", "CWRelation", "CWEType", "object"),
            ("read_permission", "CWEType", "RQLExpression", "subject"),
            ("to_entity", "CWAttribute", "CWEType", "object"),
            ("to_entity", "CWRelation", "CWEType", "object"),
            ("update_permission", "CWEType", "RQLExpression", "subject"),
        ],
        "CWRType": [
            ("relation_type", "CWAttribute", "CWRType", "object"),
            ("relation_type", "CWRelation", "CWRType", "object"),
        ],
        "CWRelation": [
            ("add_permission", "CWRelation", "RQLExpression", "subject"),
            ("constrained_by", "CWRelation", "CWConstraint", "subject"),
            ("delete_permission", "CWRelation", "RQLExpression", "subject"),
            ("read_permission", "CWRelation", "RQLExpression", "subject"),
        ],
        "CWComputedRType": [
            ("read_permission", "CWComputedRType", "RQLExpression", "subject")
        ],
        "CWSource": [
            ("cw_host_config_of", "CWSourceHostConfig", "CWSource", "object"),
            ("cw_import_of", "CWDataImport", "CWSource", "object"),
            ("cw_source", "Ami", "CWSource", "object"),
            ("cw_source", "BaseTransition", "CWSource", "object"),
            ("cw_source", "Bookmark", "CWSource", "object"),
            ("cw_source", "CWAttribute", "CWSource", "object"),
            ("cw_source", "CWComputedRType", "CWSource", "object"),
            ("cw_source", "CWConstraint", "CWSource", "object"),
            ("cw_source", "CWConstraintType", "CWSource", "object"),
            ("cw_source", "CWDataImport", "CWSource", "object"),
            ("cw_source", "CWEType", "CWSource", "object"),
            ("cw_source", "CWGroup", "CWSource", "object"),
            ("cw_source", "CWPermission", "CWSource", "object"),
            ("cw_source", "CWProperty", "CWSource", "object"),
            ("cw_source", "CWRType", "CWSource", "object"),
            ("cw_source", "CWRelation", "CWSource", "object"),
            ("cw_source", "CWSession", "CWSource", "object"),
            ("cw_source", "CWSource", "CWSource", "object"),
            ("cw_source", "CWSourceHostConfig", "CWSource", "object"),
            ("cw_source", "CWUniqueTogetherConstraint", "CWSource", "object"),
            ("cw_source", "CWUser", "CWSource", "object"),
            ("cw_source", "Card", "CWSource", "object"),
            ("cw_source", "Comment", "CWSource", "object"),
            ("cw_source", "EmailAddress", "CWSource", "object"),
            ("cw_source", "ExternalUri", "CWSource", "object"),
            ("cw_source", "FakeFile", "CWSource", "object"),
            ("cw_source", "Note", "CWSource", "object"),
            ("cw_source", "Personne", "CWSource", "object"),
            ("cw_source", "Produit", "CWSource", "object"),
            ("cw_source", "RQLExpression", "CWSource", "object"),
            ("cw_source", "Reference", "CWSource", "object"),
            ("cw_source", "Service", "CWSource", "object"),
            ("cw_source", "Societe", "CWSource", "object"),
            ("cw_source", "State", "CWSource", "object"),
            ("cw_source", "StateFull", "CWSource", "object"),
            ("cw_source", "SubNote", "CWSource", "object"),
            ("cw_source", "SubWorkflowExitPoint", "CWSource", "object"),
            ("cw_source", "Tag", "CWSource", "object"),
            ("cw_source", "TrInfo", "CWSource", "object"),
            ("cw_source", "Transition", "CWSource", "object"),
            ("cw_source", "Usine", "CWSource", "object"),
            ("cw_source", "Ville", "CWSource", "object"),
            ("cw_source", "Workflow", "CWSource", "object"),
            ("cw_source", "WorkflowTransition", "CWSource", "object"),
        ],
        "CWUser": [
            ("for_user", "CWProperty", "CWUser", "object"),
            ("use_email", "CWUser", "EmailAddress", "subject"),
            ("wf_info_for", "TrInfo", "CWUser", "object"),
        ],
        "Comment": [
            ("comments", "Comment", "Comment", "object"),
        ],
        "Personne": [
            ("comments", "Comment", "Personne", "object"),
        ],
        "StateFull": [("wf_info_for", "TrInfo", "StateFull", "object")],
        "Transition": [("condition", "Transition", "RQLExpression", "subject")],
        "Workflow": [
            ("state_of", "State", "Workflow", "object"),
            ("transition_of", "BaseTransition", "Workflow", "object"),
            ("transition_of", "Transition", "Workflow", "object"),
            ("transition_of", "WorkflowTransition", "Workflow", "object"),
        ],
        "WorkflowTransition": [
            ("condition", "WorkflowTransition", "RQLExpression", "subject"),
            (
                "subworkflow_exit",
                "WorkflowTransition",
                "SubWorkflowExitPoint",
                "subject",
            ),
        ],
    }

    def test_composite_entities(self):
        schema = self.vreg.schema
        self.assertEqual(
            sorted(self.composites),
            [
                eschema.type
                for eschema in sorted(schema.entities())
                if eschema.is_composite
            ],
        )
        for etype in self.composites:
            with self.subTest(etype=etype):
                self.assertEqual(
                    self.composites[etype],
                    sorted(
                        [
                            (r.relation_type.type, r.subject.type, r.object.type, role)
                            for r, role in schema[etype].composite_rdef_roles
                        ]
                    ),
                )


class TypescriptSchemaExporterTC(TestCase):
    def test_contains_json_export(self):
        exporter = TypescriptSchemaExporter("TestSchema")
        json_exporter = JSONSchemaExporter()

        json_export = json_exporter.export(schema)
        typescript_export = exporter.export(schema)

        assert json_export in typescript_export

    def test_default_name_generates_default_export(self):
        exporter = TypescriptSchemaExporter("default")

        export = exporter.export(schema)

        assert "export default" in export

    def test_custom_name_generates_named_export(self):
        exporter = TypescriptSchemaExporter("MyTestSchema")

        export = exporter.export(schema)

        assert "export interface MyTestSchema" in export


class JSONSchemaExporterTC(TestCase):
    def setUp(self):
        self.json_exporter = JSONSchemaExporter()
        self.schema_json = json.loads(self.json_exporter.export(schema))

    def test_generates_valid_json(self):
        assert "entities" in self.schema_json
        assert "relations_definitions" in self.schema_json

    def test_same_number_of_entities_as_in_schema(self):
        entities_count = len(schema.entities())

        assert len(self.schema_json["entities"]) == entities_count

    def test_same_number_of_relations_definitions_as_in_schema(self):
        relations_definitions_count = sum(
            len(rel.relation_definitions) for rel in schema.relations()
        )

        assert (
            len(self.schema_json["relations_definitions"])
            == relations_definitions_count
        )

    def test_entity(self):
        personne = next(
            e for e in self.schema_json["entities"] if e["type"] == "Personne"
        )

        assert personne == {
            "type": "Personne",
            "description": "A person with a name",
            "final": False,
        }

    def test_final_entity(self):
        string = next(e for e in self.schema_json["entities"] if e["type"] == "String")

        assert string == {
            "type": "String",
            "description": "",
            "final": True,
        }

    def test_attribute_definition(self):
        nom = next(
            e for e in self.schema_json["relations_definitions"] if e["type"] == "nom"
        )

        assert nom == {
            "type": "nom",
            "subject": "Personne",
            "object": "String",
            "default": None,
            "cardinality": "?1",
            "description": "",
            "final": True,
            "constraints": [],
            "options": {},
        }

    def test_relation_definition(self):
        travaille = next(
            e
            for e in self.schema_json["relations_definitions"]
            if e["type"] == "travaille"
        )

        assert travaille == {
            "type": "travaille",
            "subject": "Personne",
            "object": "Societe",
            "cardinality": "**",
            "description": "",
            "final": False,
            "constraints": [],
            "options": {},
        }

    def test_symetric_relation(self):
        rel1 = {
            "type": "colabore_avec",
            "description": "",
            "final": False,
            "subject": "Personne",
            "object": "Societe",
            "cardinality": "**",
            "constraints": [],
            "options": {"symmetric": True},
        }
        rel2 = {
            "type": "colabore_avec",
            "description": "",
            "final": False,
            "subject": "Societe",
            "object": "Personne",
            "cardinality": "**",
            "constraints": [],
            "options": {"symmetric": True},
        }
        self.assertIn(rel1, self.schema_json["relations_definitions"])
        self.assertIn(rel2, self.schema_json["relations_definitions"])


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""server side objects"""

import os.path as osp
import importlib


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)
    global URL_MAPPING
    URL_MAPPING = {}
    if vreg.config.apphome:
        url_mapping_file = osp.join(vreg.config.apphome, "urlmapping.py")
        if osp.exists(url_mapping_file):
            module = importlib.import_module(url_mapping_file)
            URL_MAPPING = getattr(module, "URL_MAPPING")
            vreg.info("using url mapping %s from %s", URL_MAPPING, url_mapping_file)



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""some hooks and views to handle supervising of any data changes"""

from cubicweb import UnknownEid
from cubicweb.mail import format_mail
from cubicweb.predicates import none_rset
from cubicweb.schema import display_name
from cubicweb.server.hook import SendMailOp
from cubicweb.sobjects.notification import NotificationView
from cubicweb.utils import UStringIO, format_and_escape_string


def filter_changes(changes):
    """
    * when an entity has been deleted:
      * don't show deletion of its relations
      * don't show related TrInfo deletion if any
    * when an entity has been added don't show owned_by relation addition
    * don't show new TrInfo entities if any
    """
    # first build an index of changes
    index = {}
    added, deleted = set(), set()
    for change in changes[:]:
        event, changedescr = change
        if event == "add_entity":
            entity = changedescr.entity
            added.add(entity.eid)
            if entity.e_schema == "TrInfo":
                changes.remove(change)
                event = "change_state"
                change = (
                    event,
                    (entity.wf_info_for[0], entity.from_state[0], entity.to_state[0]),
                )
                changes.append(change)
        elif event == "delete_entity":
            deleted.add(changedescr[0])
        index.setdefault(event, set()).add(change)
    for key in ("delete_relation", "add_relation"):
        for change in index.get(key, {}).copy():
            if change[1].rtype == "in_state":
                index[key].remove(change)
    # filter changes
    for eid in added:
        try:
            for change in index["add_relation"].copy():
                changedescr = change[1]
                # skip meta-relations which are set automatically
                # XXX generate list below using rtags (category = 'generated')
                if (
                    changedescr.rtype
                    in (
                        "created_by",
                        "owned_by",
                        "is",
                        "is_instance_of",
                        "from_state",
                        "to_state",
                        "by_transition",
                        "wf_info_for",
                    )
                    and changedescr.eidfrom == eid
                ):
                    index["add_relation"].remove(change)
        except KeyError:
            break
    for eid in deleted:
        try:
            for change in index["delete_relation"].copy():
                if change[1].eidfrom == eid:
                    index["delete_relation"].remove(change)
                elif change[1].eidto == eid:
                    index["delete_relation"].remove(change)
                    if change[1].rtype == "wf_info_for":
                        for change_ in index["delete_entity"].copy():
                            if change_[1].eidfrom == change[1].eidfrom:
                                index["delete_entity"].remove(change_)
        except KeyError:
            break
    for change in changes:
        event, changedescr = change
        if change in index[event]:
            yield change


class SupervisionEmailView(NotificationView):
    """view implementing the email API for data changes supervision notification"""

    __regid__ = "supervision_notif"
    __select__ = none_rset()

    def recipients(self):
        return self._cw.vreg.config["supervising-addrs"]

    def subject(self):
        return (
            self._cw._("[%s supervision] changes summary") % self._cw.vreg.config.appid
        )

    def render(self, changes):
        user = self._cw.user
        stream = UStringIO()
        stream.write(
            self._cw._("user %s has made the following change(s):\n\n"), user.login
        )
        for event, changedescr in filter_changes(changes):
            stream.write("* ")
            template, context = getattr(self, event)(changedescr)
            stream.write(format_and_escape_string(template, context, escape=True))
            stream.write("\n\n")
        return stream.getvalue()

    def _entity_context(self, entity):
        return {
            "eid": entity.eid,
            "etype": entity.dc_type().lower(),
            "title": entity.dc_title(),
            "url": entity.absolute_url(),
        }

    def add_entity(self, changedescr):
        return (
            "added %(etype)s #%(eid)s (%(title)s)\n  %(url)s",
            self._entity_context(changedescr.entity),
        )

    def update_entity(self, changedescr):
        return (
            "updated %(etype)s #%(eid)s (%(title)s)\n  %(url)s",
            self._entity_context(changedescr.entity),
        )

    def delete_entity(self, args):
        eid, etype, title = args
        return "deleted %(etype)s #%(eid)s (%(title)s)", {
            "eid": eid,
            "etype": display_name(self._cw, etype).lower(),
            "title": title,
        }

    def change_state(self, args):
        entity, fromstate, tostate = args
        context = self._entity_context(entity)
        context["fromstate"] = fromstate.name
        context["tostate"] = tostate.name
        return (
            "changed state of %(etype)s #%(eid)s (%(title)s)\n"
            "  from state %(fromstate)s to state %(tostate)s\n"
            "  %(url)s",
            context,
        )

    def _relation_context(self, changedescr):
        cnx = self._cw

        def describe(eid):
            try:
                return cnx._(cnx.entity_type(eid)).lower()
            except UnknownEid:
                # may occurs when an entity has been deleted from an external
                # source and we're cleaning its relation
                return cnx._("unknown external entity")

        eidfrom, rtype, eidto = (
            changedescr.eidfrom,
            changedescr.rtype,
            changedescr.eidto,
        )
        return {
            "rtype": cnx._(rtype),
            "eidfrom": eidfrom,
            "frometype": describe(eidfrom),
            "eidto": eidto,
            "toetype": describe(eidto),
        }

    def add_relation(self, changedescr):
        return (
            "added relation %(rtype)s from %(frometype)s #%(eidfrom)s to %(toetype)s"
            " #%(eidto)s",
            self._relation_context(changedescr),
        )

    def delete_relation(self, changedescr):
        return (
            "deleted relation %(rtype)s from %(frometype)s #%(eidfrom)s to %(toetype)s"
            " #%(eidto)s",
            self._relation_context(changedescr),
        )


class SupervisionMailOp(SendMailOp):
    """special send email operation which should be done only once for a bunch
    of changes
    """

    def _get_view(self):
        return self.cnx.vreg["notification-views"].select("supervision_notif", self.cnx)

    def _prepare_email(self):
        cnx = self.cnx
        config = cnx.vreg.config
        uinfo = {"email": config["sender-addr"], "name": config["sender-name"]}
        view = self._get_view()
        content = view.render(changes=cnx.transaction_data.get("pendingchanges"))
        recipients = view.recipients()
        msg = format_mail(uinfo, recipients, content, view.subject(), config=config)
        self.to_send = [(msg, recipients)]

    def postcommit_event(self):
        self._prepare_email()
        SendMailOp.postcommit_event(self)



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
import os
import sys
import unittest
from io import StringIO
from os.path import join
from unittest.mock import patch, MagicMock

import unittest_cwconfig
from logilab.common.clcommands import CommandLine

from cubicweb import utils, server
from cubicweb.__pkginfo__ import version as cw_version
from cubicweb.cwconfig import CubicWebConfiguration as cwcfg
from cubicweb.cwctl import ListCommand, InstanceCommand
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.migractions import ServerMigrationHelper


class CubicWebCtlTC(unittest.TestCase):
    setUpClass = unittest_cwconfig.CubicWebConfigurationTC.setUpClass
    tearDownClass = unittest_cwconfig.CubicWebConfigurationTC.tearDownClass

    def setUp(self):
        self.stream = StringIO()
        sys.stdout = self.stream

    def tearDown(self):
        sys.stdout = sys.__stdout__

    @patch(
        (
            "importlib_metadata.entry_points"
            if sys.version_info < (3, 10)
            else "importlib.metadata.entry_points"
        ),
        side_effect=unittest_cwconfig.iter_entry_points,
    )
    def test_list(self, mock_iter_entry_points):
        ListCommand(None).run([])
        stream_result = self.stream.getvalue()
        self.assertNotIn("cubicweb_", stream_result)
        sys.stderr.write(stream_result)
        mock_iter_entry_points.assert_called_once_with(group="cubicweb.cubes")


class CubicWebShellTC(CubicWebTC):
    def test_process_script_args_context(self):
        repo = self.repo
        with self.admin_access.repo_cnx() as cnx:
            mih = ServerMigrationHelper(
                None,
                repo=repo,
                cnx=cnx,
                interactive=False,
                # hack so it don't try to load fs schema
                schema=1,
            )
            scripts = {
                "script1.py": list(),
                "script2.py": ["-v"],
                "script3.py": ["-vd", "-f", "FILE.TXT"],
            }
            mih.cmd_process_script(
                join(self.datadir, "scripts", "script1.py"), funcname=None
            )
            for script, args in scripts.items():
                scriptname = os.path.join(self.datadir, "scripts", script)
                self.assertTrue(os.path.exists(scriptname))
                mih.cmd_process_script(scriptname, None, scriptargs=args)


class _TestCommand(InstanceCommand):
    "I need some doc"

    name = "test"
    actionverb = "failtested"

    def test_instance(self, appid):
        pass


class _TestFailCommand(InstanceCommand):
    "I need some doc"

    name = "test_fail"
    actionverb = "tested"

    def test_fail_instance(self, appid):
        raise Exception()


class InstanceCommandTest(unittest.TestCase):
    def setUp(self):
        self.CWCTL = CommandLine(
            "cubicweb-ctl",
            "The CubicWeb swiss-knife.",
            version=cw_version,
            check_duplicated_command=False,
        )
        cwcfg.load_cwctl_plugins()
        self.CWCTL.register(_TestCommand)
        self.CWCTL.register(_TestFailCommand)

        self.fake_config = MagicMock()
        self.fake_config.global_set_option = MagicMock()

        # pretend that this instance exists
        config_patcher = patch.object(
            cwcfg, "config_for", return_value=self.fake_config
        )
        config_patcher.start()
        self.addCleanup(config_patcher.stop)

    @patch.object(_TestCommand, "test_instance", return_value=0)
    def test_getting_called(self, test_instance):
        with self.assertRaises(SystemExit) as cm:
            self.CWCTL.run(["test", "some_instance"])
        self.assertEqual(cm.exception.code, 0)
        test_instance.assert_called_with("some_instance")

    @patch.object(utils, "get_pdb")
    def test_pdb_not_called(self, get_pdb):
        # CWCTL will finish the program after that
        with self.assertRaises(SystemExit) as cm:
            self.CWCTL.run(["test", "some_instance"])
        self.assertEqual(cm.exception.code, 0)

        get_pdb.assert_not_called()

    @patch.object(utils, "get_pdb")
    def test_pdb_called(self, get_pdb):
        post_mortem = get_pdb.return_value.post_mortem
        with self.assertRaises(SystemExit) as cm:
            self.CWCTL.run(["test_fail", "some_instance", "--pdb"])
        self.assertEqual(cm.exception.code, 8)

        get_pdb.assert_called_once()
        post_mortem.assert_called_once()

        # we want post_mortem to actually receive the traceback
        self.assertNotEqual(post_mortem.call_args, ((None,),))

    @patch.dict(sys.modules, ipdb=MagicMock())
    def test_ipdb_selected_and_called(self):
        ipdb = sys.modules["ipdb"]
        with self.assertRaises(SystemExit) as cm:
            self.CWCTL.run(["test_fail", "some_instance", "--pdb"])
        self.assertEqual(cm.exception.code, 8)

        ipdb.post_mortem.assert_called_once()

    @patch.object(_TestFailCommand, "test_fail_instance", side_effect=SystemExit(42))
    def test_respect_return_error_code(self, test_fail_instance):
        with self.assertRaises(SystemExit) as cm:
            self.CWCTL.run(["test_fail", "some_instance"])
        self.assertEqual(cm.exception.code, 42)

        test_fail_instance.assert_called_once()

    @patch.object(_TestFailCommand, "test_fail_instance", side_effect=KeyboardInterrupt)
    def test_error_code_keyboardinterupt_2(self, test_fail_instance):
        with self.assertRaises(SystemExit) as cm:
            self.CWCTL.run(["test_fail", "some_instance"])
        self.assertEqual(cm.exception.code, 2)

        test_fail_instance.assert_called_once()

    def test_set_loglevel(self):
        LOG_LEVELS = ("debug", "info", "warning", "error")

        for log_level in LOG_LEVELS:
            with self.assertRaises(SystemExit) as cm:
                self.CWCTL.run(["test", "some_instance", "--loglevel", log_level])
            self.assertEqual(cm.exception.code, 0)

            self.fake_config.global_set_option.assert_called_with(
                "log-threshold", log_level.upper()
            )

    @patch.object(server, "DEBUG", 0)
    def test_set_dblevel(self):
        DBG_FLAGS = ("RQL", "SQL", "REPO", "HOOKS", "OPS", "SEC", "MORE")

        total_value = 0

        for dbg_flag in DBG_FLAGS:
            with self.assertRaises(SystemExit) as cm:
                self.CWCTL.run(["test", "some_instance", "--dbglevel", dbg_flag])
            self.assertEqual(cm.exception.code, 0)

            total_value += getattr(server, f"DBG_{dbg_flag}")
            self.assertEqual(total_value, server.DEBUG)


if __name__ == "__main__":
    unittest.main()



from logilab.database import get_db_helper
from yams import register_base_type

_NUMERIC_PARAMETERS = {"scale": 0, "precision": None}
register_base_type("Numeric", _NUMERIC_PARAMETERS)

# Add the datatype to the helper mapping
pghelper = get_db_helper("postgres")


def pg_numeric_sqltype(rdef):
    """Return a PostgreSQL column type corresponding to rdef"""
    return f"numeric({rdef.precision}, {rdef.scale})"


pghelper.TYPE_MAPPING["Numeric"] = pg_numeric_sqltype



# copyright 2015 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from logilab.common import tempattr

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.sources.native import FTIndexEntityOp


class NativeSourceTC(CubicWebTC):
    def test_index_entity_consider_do_fti(self):
        source = self.repo.system_source
        with tempattr(source, "do_fti", False):
            with self.admin_access.repo_cnx() as cnx:
                # when do_fti is set to false, call to index_entity (as may be done from hooks)
                # should have no effect
                source.index_entity(cnx, cnx.user)
                self.assertNotIn(
                    cnx.user.eid, FTIndexEntityOp.get_instance(cnx).get_data()
                )


if __name__ == "__main__":
    from logilab.common.testlib import unittest_main

    unittest_main()



# copyright 2013-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <https://www.gnu.org/licenses/>.
"""Massive store test case"""

import psycopg2

from cubicweb.dataimport import stores
from cubicweb.dataimport.csv import ucsvreader
from cubicweb.dataimport.massive_store import MassiveObjectStore, PGHelper
from cubicweb.dataimport.test import test_stores
from cubicweb.devtools import startpgcluster, stoppgcluster
from cubicweb.devtools import testlib
from cubicweb.devtools.apptest_config import PostgresApptestConfiguration
from cubicweb.server.schema2sql import build_index_name


def setUpModule():
    startpgcluster(__file__)


def tearDownModule(*args):
    stoppgcluster(__file__)


def all_indexes(cnx):
    crs = cnx.system_sql("SELECT indexname FROM pg_indexes")
    return {r[0] for r in crs.fetchall()}


class MassiveObjectStoreWithCustomMDGenStoreTC(
    test_stores.NoHookRQLObjectStoreWithCustomMDGenStoreTC
):
    configcls = PostgresApptestConfiguration

    def store_impl(self, cnx):
        source = cnx.create_entity("CWSource", type="datafeed", name="test", url="test")
        cnx.commit()
        metagen = stores.MetadataGenerator(
            cnx, source=cnx.repo.source_by_eid(source.eid)
        )
        return MassiveObjectStore(cnx, metagen=metagen)


class MassImportSimpleTC(testlib.CubicWebTC):
    configcls = PostgresApptestConfiguration
    appid = "data-massimport"

    def cast(self, _type, value):
        try:
            return _type(value)
        except ValueError:
            return None

    def push_geonames_data(self, dumpname, store):
        # Push timezones
        cnx = store._cnx
        for code, gmt, dst, raw_offset in ucsvreader(
            open(self.datapath("timeZones.txt"), "rb"), delimiter="\t"
        ):
            cnx.create_entity(
                "TimeZone",
                code=code,
                gmt=float(gmt),
                dst=float(dst),
                raw_offset=float(raw_offset),
            )
        timezone_code = dict(cnx.execute("Any C, X WHERE X is TimeZone, X code C"))
        # Push data
        for ind, infos in enumerate(
            ucsvreader(open(dumpname, "rb"), delimiter="\t", ignore_errors=True)
        ):
            latitude = self.cast(float, infos[4])
            longitude = self.cast(float, infos[5])
            population = self.cast(int, infos[14])
            elevation = self.cast(int, infos[15])
            gtopo = self.cast(int, infos[16])
            feature_class = infos[6]
            if len(infos[6]) != 1:
                feature_class = None
            entity = {
                "name": infos[1],
                "asciiname": infos[2],
                "alternatenames": infos[3],
                "latitude": latitude,
                "longitude": longitude,
                "feature_class": feature_class,
                "alternate_country_code": infos[9],
                "admin_code_3": infos[12],
                "admin_code_4": infos[13],
                "population": population,
                "elevation": elevation,
                "gtopo30": gtopo,
                "timezone": timezone_code.get(infos[17]),
                "cwuri": f"http://sws.geonames.org/{int(infos[0])}/",
                "geonameid": int(infos[0]),
            }
            store.prepare_insert_entity("Location", **entity)

    def test_autoflush_metadata(self):
        with self.admin_access.repo_cnx() as cnx:
            crs = cnx.system_sql(
                "SELECT * FROM entities WHERE type=%(t)s", {"t": "Location"}
            )
            self.assertEqual(len(crs.fetchall()), 0)
            store = MassiveObjectStore(cnx)
            store.prepare_insert_entity("Location", name="toto")
            store.flush()
            store.commit()
            store.finish()
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            crs = cnx.system_sql(
                "SELECT * FROM entities WHERE type=%(t)s", {"t": "Location"}
            )
            self.assertEqual(len(crs.fetchall()), 1)

    def test_massimport_etype_metadata(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            timezone_eid = store.prepare_insert_entity("TimeZone", code="12")
            store.prepare_insert_entity("Location", timezone=timezone_eid)
            store.flush()
            store.commit()
            store.finish()
            eid, etname = cnx.execute(
                "Any X, TN WHERE X timezone TZ, X is T, " "T name TN"
            )[0]
            self.assertEqual(cnx.entity_from_eid(eid).cw_etype, etname)

    def test_index_not_dropped_by_init(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)  # noqa
            cnx.commit()
            indexes = all_indexes(cnx)
            self.assertIn("entities_pkey", indexes)
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from", "eid_to"], "key_"),
                indexes,
            )
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from"], "idx_"), indexes
            )

    def test_drop_index_recreation(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)

            store._drop_metadata_constraints()
            indexes = all_indexes(cnx)
            self.assertIn("entities_pkey", indexes)
            self.assertNotIn(
                build_index_name("owned_by_relation", ["eid_from", "eid_to"], "key_"),
                indexes,
            )
            self.assertNotIn(
                build_index_name("owned_by_relation", ["eid_from"], "idx_"), indexes
            )

            store.finish()
            indexes = all_indexes(cnx)
            self.assertIn("entities_pkey", indexes)
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from", "eid_to"], "key_"),
                indexes,
            )
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from"], "idx_"), indexes
            )

    def test_consider_metagen(self):
        """Ensure index on owned_by is not deleted if we don't consider this metadata."""
        with self.admin_access.repo_cnx() as cnx:
            metagen = stores.MetadataGenerator(cnx, meta_skipped=("owned_by",))
            store = MassiveObjectStore(cnx, metagen=metagen)
            store._drop_metadata_constraints()

            indexes = all_indexes(cnx)
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from", "eid_to"], "key_"),
                indexes,
            )
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from"], "idx_"), indexes
            )

    def test_eids_seq_range(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx, eids_seq_range=1000)
            store.restart_eid_sequence(50000)
            store.prepare_insert_entity("Location", name="toto")
            store.flush()
            cnx.commit()
        with self.admin_access.repo_cnx() as cnx:
            crs = cnx.system_sql("SELECT * FROM entities_id_seq")
            self.assertGreater(crs.fetchone()[0], 50000)

    def test_eid_entity(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx, eids_seq_range=1000)
            store.restart_eid_sequence(50000)
            eid = store.prepare_insert_entity("Location", name="toto")
            store.flush()
            self.assertGreater(eid, 50000)

    def test_eid_entity_2(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            store.restart_eid_sequence(50000)
            eid = store.prepare_insert_entity("Location", name="toto", eid=10000)
            store.flush()
        self.assertEqual(eid, 10000)

    @staticmethod
    def get_db_descr(cnx):
        pgh = PGHelper(cnx)
        all_tables = cnx.system_sql(
            """
SELECT table_name
FROM information_schema.tables
where table_schema = %(s)s""",
            {"s": pgh.pg_schema},
        ).fetchall()
        all_tables_descr = {}
        for (tablename,) in all_tables:
            all_tables_descr[tablename] = set(pgh.table_indexes(tablename)).union(
                set(pgh.table_constraints(tablename))
            )
        return all_tables_descr

    def test_identical_schema(self):
        with self.admin_access.repo_cnx() as cnx:
            init_descr = self.get_db_descr(cnx)
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            store.prepare_insert_entity("Location", name="toto")
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            final_descr = self.get_db_descr(cnx)
        self.assertEqual(init_descr, final_descr)

    def test_simple_insert(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            self.push_geonames_data(self.datapath("geonames.csv"), store)
            store.flush()
            store.commit()
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any X WHERE X is Location")
            self.assertEqual(len(rset), 4000)
            rset = cnx.execute("Any X WHERE X is Location, X timezone T")
            self.assertEqual(len(rset), 4000)

    def test_simple_update(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            self.push_geonames_data(self.datapath("geonames.csv"), store)
            store.flush()
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any X LIMIT 2000 WHERE X is Location")
            self.assertEqual(len(rset), 2000)
            store = MassiveObjectStore(cnx)
            for (eid,) in rset:
                store.prepare_update_entity("Location", eid, name="Poulet")
            store.flush()
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any X WHERE X is Location, X name 'Poulet'")
            self.assertEqual(len(rset), 2000)

    def test_simple_insert_no_drop_index_nor_constraints(self):
        """The same test as test_simple_insert, but without dropping indexes
        nor constraints.
        """
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx, drop=False)
            self.push_geonames_data(self.datapath("geonames.csv"), store)
            store.flush()
            store.commit()
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any X WHERE X is Location")
            self.assertEqual(len(rset), 4000)
            rset = cnx.execute("Any X WHERE X is Location, X timezone T")
            self.assertEqual(len(rset), 4000)

    def test_simple_update_no_drop_index_nor_constraints(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx, drop=False)
            self.push_geonames_data(self.datapath("geonames.csv"), store)
            store.flush()
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any X LIMIT 2000 WHERE X is Location")
            self.assertEqual(len(rset), 2000)
            store = MassiveObjectStore(cnx)
            for (eid,) in rset:
                store.prepare_update_entity("Location", eid, name="Poulet")
            store.flush()
            store.finish()
        with self.admin_access.repo_cnx() as cnx:
            rset = cnx.execute("Any X WHERE X is Location, X name 'Poulet'")
            self.assertEqual(len(rset), 2000)

    def test_index_building(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            self.push_geonames_data(self.datapath("geonames.csv"), store)
            store.flush()

            # Check index
            indexes = all_indexes(cnx)
            self.assertIn("entities_pkey", indexes)
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from", "eid_to"], "key_"),
                indexes,
            )
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from"], "idx_"), indexes
            )

            # Cleanup -> index
            store.finish()

            # Check index again
            indexes = all_indexes(cnx)
            self.assertIn("entities_pkey", indexes)
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from", "eid_to"], "key_"),
                indexes,
            )
            self.assertIn(
                build_index_name("owned_by_relation", ["eid_from"], "idx_"), indexes
            )

    def test_multiple_insert(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            store.prepare_insert_entity("Location", name="toto")
            store.finish()
            store = MassiveObjectStore(cnx)
            store.prepare_insert_entity("Location", name="toto")
            store.finish()

    def test_delete_metatable_on_integrity_error(self):
        with self.admin_access.repo_cnx() as cnx:
            store = MassiveObjectStore(cnx)
            store.prepare_insert_entity("TimeZone")
            store.flush()
            store.commit()
            with self.assertRaises(psycopg2.IntegrityError):
                store.finish()
            self.assertNotIn("cwmassive_initialized", set(self.get_db_descr(cnx)))

    def test_massive_store_table_deleted_on_failure(self):
        """assert that the 'store.finish' has cleaned the automatically created
        table if an exception is raised during the import"""

        class FakeException(Exception):
            pass

        with self.admin_access.repo_cnx() as cnx:
            try:
                with MassiveObjectStore(cnx) as store:
                    store.prepare_insert_entity("Location", name="toto")
                    store.flush()
                    raise FakeException
            except FakeException:
                # assert no location has been inserted
                nb_locations = cnx.system_sql(
                    """
                       SELECT COUNT(*) FROM cw_location;
                    """
                ).fetchone()[0]
                self.assertEqual(nb_locations, 0)

                # assert cwmassive_initialized table has been deleted
                result = cnx.system_sql(
                    """
                    SELECT EXISTS (
                       SELECT FROM information_schema.tables
                       WHERE table_name = 'cwmassive_initialized'
                    );
                    """
                )
                self.assertFalse(result.fetchone()[0])

                # assert there is no temporary table left
                result = cnx.system_sql(
                    """
                    SELECT EXISTS (
                       SELECT FROM information_schema.tables
                       WHERE table_name ILIKE 'cw_location_%'
                    );
                    """
                )
                self.assertFalse(result.fetchone()[0])


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2004-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of yams.
#
# yams is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# yams is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with yams. If not, see <https://www.gnu.org/licenses/>.
from yams.buildobjs import (
    EntityType,
    RelationType,
    SubjectRelation,
    Int,
    String,
    Boolean,
)
from yams.constraints import SizeConstraint, UniqueConstraint

from . import RESTRICTED_RTYPE_PERMS


class State(EntityType):
    """used to associate simple states to an entity
    type and/or to define workflows
    """

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            "users",
        ),
        "delete": (
            "managers",
            "owners",
        ),
        "update": (
            "managers",
            "owners",
        ),
    }

    # attributes
    eid = Int(required=True, uid=True)
    name = String(
        required=True,
        indexed=True,
        internationalizable=True,
        constraints=[SizeConstraint(max=256)],
    )
    description = String(fulltextindexed=True)
    # relations
    state_of = SubjectRelation("Eetype", cardinality="+*")
    next_state = SubjectRelation("State", cardinality="**")


class state_of(RelationType):
    """link a state to one or more entity type"""

    __permissions__ = RESTRICTED_RTYPE_PERMS


class next_state(RelationType):
    """define a workflow by associating a state to possible following states"""

    __permissions__ = RESTRICTED_RTYPE_PERMS


class initial_state(RelationType):
    """indicate which state should be used by default when an entity using states
    is created
    """

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": (
            "managers",
            "users",
        ),
        "delete": (
            "managers",
            "users",
        ),
    }
    subject = "Eetype"
    object = "State"
    cardinality = "?*"
    inlined = True


class Eetype(EntityType):
    """define an entity type, used to build the application schema"""

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),
        "add": ("managers",),
        "delete": ("managers",),
        "update": (
            "managers",
            "owners",
        ),
    }
    name = String(
        required=True,
        indexed=True,
        internationalizable=True,
        constraints=[UniqueConstraint(), SizeConstraint(max=64)],
    )
    description = String(fulltextindexed=True)
    meta = Boolean()
    final = Boolean()



# copyright 2003-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Postgres specific store"""

import os.path as osp
import pickle
import warnings
from collections import defaultdict
from datetime import date, datetime, time
from io import StringIO
from time import asctime

from cubicweb.server.sqlutils import SQL_PREFIX
from cubicweb.utils import make_uid


def _execmany_thread_not_copy_from(
    cu, statement, data, table=None, columns=None, encoding="utf-8"
):
    """Execute thread without copy from"""
    cu.executemany(statement, data)


def _execmany_thread_copy_from(cu, statement, data, table, columns, encoding="utf-8"):
    """Execute thread with copy from"""
    try:
        buf = _create_copyfrom_buffer(data, columns, encoding=encoding)
    except ValueError:
        _execmany_thread_not_copy_from(cu, statement, data)
    else:
        if columns is None:
            cu.copy_from(buf, table, null="NULL")
        else:
            cu.copy_from(buf, table, null="NULL", columns=columns)


def _execmany_thread(
    sql_connect,
    statements,
    dump_output_dir=None,
    support_copy_from=True,
    encoding="utf-8",
):
    """
    Execute sql statement. If 'INSERT INTO', try to use 'COPY FROM' command,
    or fallback to execute_many.
    """
    if support_copy_from:
        execmany_func = _execmany_thread_copy_from
    else:
        execmany_func = _execmany_thread_not_copy_from
    cnx = sql_connect()
    cu = cnx.cursor()
    try:
        for statement, data in statements:
            table = None
            columns = None
            try:
                if not statement.startswith("INSERT INTO"):
                    cu.executemany(statement, data)
                    continue
                table = statement.split()[2]
                if isinstance(data[0], (tuple, list)):
                    columns = None
                else:
                    columns = list(data[0])
                execmany_func(cu, statement, data, table, columns, encoding)
            except Exception:
                print(f"unable to copy data into table {table}")
                # Error in import statement, save data in dump_output_dir
                if dump_output_dir is not None:
                    pdata = {
                        "data": data,
                        "statement": statement,
                        "time": asctime(),
                        "columns": columns,
                    }
                    filename = make_uid()
                    try:
                        with open(
                            osp.join(dump_output_dir, f"{filename}.pickle"), "wb"
                        ) as fobj:
                            pickle.dump(pdata, fobj)
                    except OSError:
                        print(
                            "ERROR while pickling in",
                            dump_output_dir,
                            filename + ".pickle",
                        )
                cnx.rollback()
                raise
    finally:
        cnx.commit()
        cu.close()


def _copyfrom_buffer_convert_None(value, **opts):
    '''Convert None value to "NULL"'''
    return "NULL"


def _copyfrom_buffer_convert_number(value, **opts):
    """Convert a number into its string representation"""
    return str(value)


def _copyfrom_buffer_convert_string(value, **opts):
    """Convert string value."""
    escape_chars = (("\\", "\\\\"), ("\t", "\\t"), ("\r", "\\r"), ("\n", "\\n"))
    for char, replace in escape_chars:
        value = value.replace(char, replace)
    return value


def _copyfrom_buffer_convert_date(value, **opts):
    '''Convert date into "YYYY-MM-DD"'''
    # Do not use strftime, as it yields issue with date < 1900
    # (http://bugs.python.org/issue1777412)
    return "%04d-%02d-%02d" % (value.year, value.month, value.day)


def _copyfrom_buffer_convert_datetime(value, **opts):
    '''Convert date into "YYYY-MM-DD HH:MM:SS.UUUUUU"'''
    # Do not use strftime, as it yields issue with date < 1900
    # (http://bugs.python.org/issue1777412)
    return "{} {}".format(
        _copyfrom_buffer_convert_date(value, **opts),
        _copyfrom_buffer_convert_time(value, **opts),
    )


def _copyfrom_buffer_convert_time(value, **opts):
    '''Convert time into "HH:MM:SS.UUUUUU"'''
    return "%02d:%02d:%02d.%06d" % (
        value.hour,
        value.minute,
        value.second,
        value.microsecond,
    )


# (types, converter) list.
_COPYFROM_BUFFER_CONVERTERS = [
    (type(None), _copyfrom_buffer_convert_None),
    ((int, float), _copyfrom_buffer_convert_number),
    (str, _copyfrom_buffer_convert_string),
    (datetime, _copyfrom_buffer_convert_datetime),
    (date, _copyfrom_buffer_convert_date),
    (time, _copyfrom_buffer_convert_time),
]


def _create_copyfrom_buffer(data, columns=None, **convert_opts):
    """
    Create a StringIO buffer for 'COPY FROM' command.
    Deals with Unicode, Int, Float, Date... (see ``converters``)

    :data: a sequence/dict of tuples
    :columns: list of columns to consider (default to all columns)
    :converter_opts: keyword arguements given to converters
    """
    # Create a list rather than directly create a StringIO
    # to correctly write lines separated by '\n' in a single step
    rows = []
    if columns is None:
        if isinstance(data[0], (tuple, list)):
            columns = list(range(len(data[0])))
        elif isinstance(data[0], dict):
            columns = data[0].keys()
        else:
            raise ValueError("Could not get columns: you must provide columns.")
    for row in data:
        # Iterate over the different columns and the different values
        # and try to convert them to a correct datatype.
        # If an error is raised, do not continue.
        formatted_row = []
        for col in columns:
            try:
                value = row[col]
            except KeyError:
                warnings.warn(
                    f"Column {col} is not accessible in row {row}", RuntimeWarning
                )
                # XXX 'value' set to None so that the import does not end in
                # error.
                # Instead, the extra keys are set to NULL from the
                # database point of view.
                value = None
            for types, converter in _COPYFROM_BUFFER_CONVERTERS:
                if isinstance(value, types):
                    value = converter(value, **convert_opts)
                    assert isinstance(value, str)
                    break
            else:
                raise ValueError(f"Unsupported value type {type(value)}")
            # We push the value to the new formatted row
            # if the value is not None and could be converted to a string.
            formatted_row.append(value)
        rows.append("\t".join(formatted_row))
    return StringIO("\n".join(rows))


###########################################################################
#  SQL Source #############################################################
###########################################################################


class SQLGenSourceWrapper:
    def __init__(self, system_source, schema, dump_output_dir=None):
        self.system_source = system_source
        # Explicitely backport attributes from system source
        self._storage_handler = self.system_source._storage_handler
        self.preprocess_entity = self.system_source.preprocess_entity
        self.sqlgen = self.system_source.sqlgen
        self.uri = self.system_source.uri
        self.eid = self.system_source.eid
        # Directory to write temporary files
        self.dump_output_dir = dump_output_dir
        # Allow to execute code with SQLite backend that does
        # not support (yet...) copy_from
        # XXX Should be dealt with in logilab.database
        spcfrom = system_source.dbhelper.dbapi_module.support_copy_from
        self.support_copy_from = spcfrom
        self.dbencoding = system_source.dbhelper.dbencoding
        self.init_statement_lists()
        self._inlined_rtypes_cache = {}
        self._fill_inlined_rtypes_cache(schema)
        self.schema = schema
        self.do_fti = False

    def _fill_inlined_rtypes_cache(self, schema):
        cache = self._inlined_rtypes_cache
        for eschema in schema.entities():
            for rschema in eschema.ordered_relations():
                if rschema.inlined:
                    cache[eschema.type] = SQL_PREFIX + rschema.type

    def init_statement_lists(self):
        self._sql_entities = defaultdict(list)
        self._sql_relations = {}
        self._sql_inlined_relations = {}
        self._sql_eids = defaultdict(list)
        # keep track, for each eid of the corresponding data dict
        self._sql_eid_insertdicts = {}

    def flush(self):
        print("starting flush")
        _entities_sql = self._sql_entities
        _relations_sql = self._sql_relations
        _inlined_relations_sql = self._sql_inlined_relations
        _insertdicts = self._sql_eid_insertdicts
        try:
            # try, for each inlined_relation, to find if we're also creating
            # the host entity (i.e. the subject of the relation).
            # In that case, simply update the insert dict and remove
            # the need to make the
            # UPDATE statement
            for statement, datalist in _inlined_relations_sql.items():
                new_datalist = []
                # for a given inlined relation,
                # browse each couple to be inserted
                for data in datalist:
                    keys = list(data)
                    # For inlined relations, it exists only two case:
                    # (rtype, cw_eid) or (cw_eid, rtype)
                    if keys[0] == "cw_eid":
                        rtype = keys[1]
                    else:
                        rtype = keys[0]
                    updated_eid = data["cw_eid"]
                    if updated_eid in _insertdicts:
                        _insertdicts[updated_eid][rtype] = data[rtype]
                    else:
                        # could not find corresponding insert dict, keep the
                        # UPDATE query
                        new_datalist.append(data)
                _inlined_relations_sql[statement] = new_datalist
            _execmany_thread(
                self.system_source.get_connection,
                list(self._sql_eids.items())
                + list(_entities_sql.items())
                + list(_relations_sql.items())
                + list(_inlined_relations_sql.items()),
                dump_output_dir=self.dump_output_dir,
                support_copy_from=self.support_copy_from,
                encoding=self.dbencoding,
            )
        finally:
            _entities_sql.clear()
            _relations_sql.clear()
            _insertdicts.clear()
            _inlined_relations_sql.clear()

    def add_relation(self, cnx, subject, rtype, object, inlined=False, **kwargs):
        if inlined:
            _sql = self._sql_inlined_relations
            data = {"cw_eid": subject, SQL_PREFIX + rtype: object}
            subjtype = kwargs.get("subjtype")
            if subjtype is None:
                # Try to infer it
                targets = [
                    t.type for t in self.schema.relation_schema_for(rtype).subjects()
                ]
                if len(targets) == 1:
                    subjtype = targets[0]
                else:
                    raise ValueError(
                        "You should give the subject etype for "
                        "inlined relation %s"
                        ", as it cannot be inferred: "
                        "this type is given as keyword argument "
                        "``subjtype``" % rtype
                    )
            statement = self.sqlgen.update(SQL_PREFIX + subjtype, data, ["cw_eid"])
        else:
            _sql = self._sql_relations
            data = {"eid_from": subject, "eid_to": object}
            statement = self.sqlgen.insert(f"{rtype}_relation", data)
        if statement in _sql:
            _sql[statement].append(data)
        else:
            _sql[statement] = [data]

    def add_entity(self, cnx, entity):
        with self._storage_handler(cnx, entity, "added"):
            attrs = self.preprocess_entity(entity)
            rtypes = self._inlined_rtypes_cache.get(entity.cw_etype, ())
            if isinstance(rtypes, str):
                rtypes = (rtypes,)
            for rtype in rtypes:
                if rtype not in attrs:
                    attrs[rtype] = None
            sql = self.sqlgen.insert(SQL_PREFIX + entity.cw_etype, attrs)
            self._sql_eid_insertdicts[entity.eid] = attrs
            self._append_to_entities(sql, attrs)

    def _append_to_entities(self, sql, attrs):
        self._sql_entities[sql].append(attrs)

    def _handle_insert_entity_sql(self, cnx, sql, attrs):
        self._sql_eids[sql].append(attrs)

    def _handle_is_relation_sql(self, cnx, sql, attrs):
        self._append_to_entities(sql, attrs)

    def _handle_is_instance_of_sql(self, cnx, sql, attrs):
        self._append_to_entities(sql, attrs)

    def _handle_source_relation_sql(self, cnx, sql, attrs):
        self._append_to_entities(sql, attrs)

    # add_info is _copypasted_ from the one in NativeSQLSource. We want it
    # there because it will use the _handlers of the SQLGenSourceWrapper, which
    # are not like the ones in the native source.
    def add_info(self, cnx, entity, source):
        """add type and source info for an eid into the system table"""
        # begin by inserting eid/type/source into the entities table
        attrs = {"type": entity.cw_etype, "eid": entity.eid}
        self._handle_insert_entity_sql(
            cnx, self.sqlgen.insert("entities", attrs), attrs
        )
        # insert core relations: is, is_instance_of and cw_source
        self._handle_is_relation_sql(
            cnx,
            "INSERT INTO is_relation(eid_from,eid_to) VALUES (%s,%s)",
            (entity.eid, entity.e_schema.eid),
        )
        for eschema in entity.e_schema.ancestors() + [entity.e_schema]:
            self._handle_is_relation_sql(
                cnx,
                "INSERT INTO is_instance_of_relation(eid_from,eid_to) VALUES (%s,%s)",
                (entity.eid, eschema.eid),
            )
        self._handle_is_relation_sql(
            cnx,
            "INSERT INTO cw_source_relation(eid_from,eid_to) VALUES (%s,%s)",
            (entity.eid, source.eid),
        )
        # now we can update the full text index
        if self.do_fti and self.need_fti_indexation(entity.cw_etype):
            self.index_entity(cnx, entity=entity)



# copyright 2011-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from contextlib import contextmanager
from datetime import timedelta

from yams import ValidationError
from cubicweb.dataimport.stores import NoHookRQLObjectStore, MetadataGenerator
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.server.sources import datafeed


class DataFeedTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.base_parser(cnx):
                cnx.create_entity(
                    "CWSource",
                    name="ô myfeed",
                    type="datafeed",
                    parser="testparser",
                    url="ignored",
                    config="synchronization-interval=1min",
                )
                cnx.commit()

    @contextmanager
    def base_parser(self, session):
        class AParser(datafeed.DataFeedParser):
            __regid__ = "testparser"

            def process(self, url, raise_on_error=False):
                metagenerator = MetadataGenerator(self._cw, source=self.source)
                store = NoHookRQLObjectStore(self._cw, metagenerator)
                store.prepare_insert_entity(
                    "Card",
                    cwuri="http://www.cubicweb.org/",
                    title="cubicweb.org",
                    content="the cw web site",
                )
                store.flush()
                store.commit()

        with self.temporary_appobjects(AParser):
            try:
                source = self.repo.source_by_uri("ô myfeed")
            except ValueError:
                yield
            else:
                yield source._get_parser(session)
        # vreg.unregister just pops appobjects from their regid entry,
        # completely remove the entry to ensure we have no side effect with
        # this empty entry.
        del self.vreg["parsers"][AParser.__regid__]

    def test(self):
        dfsource = self.repo.source_by_uri("ô myfeed")
        self.assertNotIn("use_cwuri_as_url", dfsource.__dict__)
        self.assertEqual(
            {"type": "datafeed", "uri": "ô myfeed", "use-cwuri-as-url": True},
            dfsource.public_config,
        )
        self.assertEqual(dfsource.use_cwuri_as_url, True)
        self.assertEqual(dfsource.latest_retrieval, None)
        self.assertEqual(dfsource.synchro_interval, timedelta(seconds=60))
        self.assertFalse(dfsource.fresh())
        # ensure source's logger name has been unormalized
        self.assertEqual(dfsource.info.__self__.name, "cubicweb.sources.o myfeed")

        with self.repo.internal_cnx() as cnx:
            with self.base_parser(cnx):
                stats = dfsource.pull_data(cnx, force=True, raise_on_error=True)
                cnx.commit()
                # test import stats
                self.assertEqual(sorted(stats), ["checked", "created", "updated"])
                entity = cnx.execute("Card X").get_entity(0, 0)
                # test imported entities
                self.assertEqual(entity.title, "cubicweb.org")
                self.assertEqual(entity.content, "the cw web site")
                self.assertEqual(entity.cwuri, "http://www.cubicweb.org/")
                self.assertEqual(entity.cw_source[0].name, "ô myfeed")
                # test repo cache keys
                self.assertEqual(self.repo._type_cache[entity.eid], "Card")

                self.assertTrue(dfsource.latest_retrieval)
                self.assertTrue(dfsource.fresh())

        # test_rename_source
        with self.admin_access.repo_cnx() as cnx:
            cnx.entity_from_eid(dfsource.eid).cw_set(name="myrenamedfeed")
            cnx.commit()
            entity = cnx.execute("Card X").get_entity(0, 0)
            self.assertEqual(entity.cwuri, "http://www.cubicweb.org/")
            self.assertEqual(entity.cw_source[0].name, "myrenamedfeed")
            self.assertEqual(self.repo._type_cache[entity.eid], "Card")

            # test_delete_source
            cnx.execute('DELETE CWSource S WHERE S name "myrenamedfeed"')
            cnx.commit()
            self.assertFalse(cnx.execute('Card X WHERE X title "cubicweb.org"'))
            self.assertFalse(cnx.execute('Any X WHERE X has_text "cubicweb.org"'))

    def test_parser_retrieve_url_local(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.base_parser(cnx) as parser:
                value = parser.retrieve_url("a string")
                self.assertEqual(200, value.getcode())
                self.assertEqual("a string", value.geturl())

    def test_update_url(self):
        dfsource = self.repo.source_by_uri("ô myfeed")
        with self.admin_access.repo_cnx() as cnx:
            cnx.entity_from_eid(dfsource.eid).cw_set(
                url="http://pouet.com\nhttp://pouet.org"
            )
            cnx.commit()
        self.assertEqual(dfsource.urls, ["ignored"])
        dfsource = self.repo.source_by_uri("ô myfeed")
        self.assertEqual(dfsource.urls, ["http://pouet.com", "http://pouet.org"])

    def test_parser_not_found(self):
        dfsource = self.repo.source_by_uri("ô myfeed")
        with self.assertLogs("cubicweb.sources.o myfeed", level="ERROR") as cm:
            with self.repo.internal_cnx() as cnx:
                stats = dfsource.pull_data(cnx, force=True)
                importlog = cnx.find("CWDataImport").one().log
        self.assertIn("failed to load parser for", cm.output[0])
        self.assertEqual(stats, {})
        self.assertIn(
            "failed to load parser for source &quot;ô myfeed&quot;", importlog
        )

    def test_bad_config(self):
        with self.admin_access.repo_cnx() as cnx:
            with self.base_parser(cnx):
                with self.assertRaises(ValidationError) as cm:
                    cnx.create_entity(
                        "CWSource",
                        name="error",
                        type="datafeed",
                        parser="testparser",
                        url="ignored",
                        config="synchronization-interval=1s",
                    )
                self.assertIn(
                    "synchronization-interval must be greater than 1 minute",
                    str(cm.exception),
                )
                cnx.rollback()

                with self.assertRaises(ValidationError) as cm:
                    cnx.create_entity(
                        "CWSource",
                        name="error",
                        type="datafeed",
                        parser="testparser",
                        url=None,
                        config="synchronization-interval=1min",
                    )
                self.assertIn("specifying an URL is mandatory", str(cm.exception))
                cnx.rollback()

                with self.assertRaises(ValidationError) as cm:
                    cnx.create_entity(
                        "CWSource",
                        name="error",
                        type="datafeed",
                        parser="testparser",
                        url="ignored",
                        config="synch-interval=1min",
                    )
                self.assertIn("unknown options synch-interval", str(cm.exception))
                cnx.rollback()


class DataFeedConfigTC(CubicWebTC):
    def test_use_cwuri_as_url_override(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.create_entity(
                "CWSource",
                name="myfeed",
                type="datafeed",
                parser="testparser",
                url="ignored",
                config="use-cwuri-as-url=no",
            )
            cnx.commit()
        dfsource = self.repo.source_by_uri("myfeed")
        self.assertEqual(dfsource.use_cwuri_as_url, False)
        self.assertEqual(
            {"type": "datafeed", "uri": "myfeed", "use-cwuri-as-url": False},
            dfsource.public_config,
        )


if __name__ == "__main__":
    import unittest

    unittest.main()



# copyright 2003-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""cubicweb post creation script, set user's workflow"""
from cubicweb import _


# insert versions
create_entity(
    "CWProperty", pkey="system.version.cubicweb", value=str(config.cubicweb_version())
)
for cube in config.cubes():
    create_entity(
        "CWProperty",
        pkey=f"system.version.{cube.lower()}",
        value=str(config.cube_version(cube)),
    )

# some entities have been added before schema entities, add their missing 'is' and
# 'is_instance_of' relations
for rtype in ("is", "is_instance_of"):
    sql(
        "INSERT INTO %s_relation "
        "SELECT X.eid, ET.cw_eid FROM entities as X, cw_CWEType as ET "
        "WHERE X.type=ET.cw_name AND NOT EXISTS("
        "      SELECT 1 from %s_relation "
        "      WHERE eid_from=X.eid AND eid_to=ET.cw_eid)" % (rtype, rtype)
    )

# user workflow
userwf = add_workflow(_("default user workflow"), "CWUser")
activated = userwf.add_state(_("activated"), initial=True)
deactivated = userwf.add_state(_("deactivated"))
userwf.add_transition(
    _("deactivate"), (activated,), deactivated, requiredgroups=("managers",)
)
userwf.add_transition(
    _("activate"), (deactivated,), activated, requiredgroups=("managers",)
)

# create anonymous user if all-in-one config and anonymous user has been specified
if hasattr(config, "anonymous_user"):
    anonlogin, anonpwd = config.anonymous_user()
    if anonlogin == session.user.login:
        print("you are using a manager account as anonymous user.")
        print("Hopefully this is not a production instance...")
    elif anonlogin:
        from cubicweb.server import create_user

        create_user(session, anonlogin, anonpwd, "guests")

# need this since we already have at least one user in the database (the default admin)
for user in rql("Any X WHERE X is CWUser").entities():
    rql(
        "SET X in_state S WHERE X eid %(x)s, S eid %(s)s",
        {"x": user.eid, "s": activated.eid},
    )

# on interactive mode, ask for level 0 persistent options
if interactive_mode:
    cfg = config.persistent_options_configuration()
    cfg.input_config(inputlevel=0)
    for section, options in cfg.options_by_section():
        for optname, optdict, value in options:
            key = f"{section}.{optname}"
            default = cfg.option_default(optname, optdict)
            # only record values differing from default
            if value != default:
                rql(
                    "INSERT CWProperty X: X pkey %(k)s, X value %(v)s",
                    {"k": key, "v": value},
                )



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
import re

from logilab.common.testlib import unittest_main

from cubicweb.devtools import BASE_URL
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.sobjects.supervising import SupervisionMailOp


class SupervisingTC(CubicWebTC):
    def setup_database(self):
        with self.admin_access.client_cnx() as cnx:
            cnx.create_entity("Card", title="une news !", content="cubicweb c'est beau")
            card = cnx.create_entity(
                "Card", title="une autre news !", content="cubicweb c'est beau"
            )
            cnx.create_entity("Bookmark", title="un signet !", path="view?vid=index")
            cnx.create_entity("Comment", content="Yo !", comments=card)
            cnx.commit()
        self.vreg.config.global_set_option("supervising-addrs", "test@logilab.fr")

    def test_supervision(self):
        # do some modification
        with self.admin_access.repo_cnx() as cnx:
            user = cnx.execute(
                'INSERT CWUser X: X login "toto", X upassword "sosafe", X in_group G '
                'WHERE G name "users"'
            ).get_entity(0, 0)
            cnx.execute("SET X last_login_time NOW WHERE X eid %(x)s", {"x": user.eid})
            cnx.execute('DELETE Card B WHERE B title "une news !"')
            cnx.execute(
                "SET X bookmarked_by U WHERE X is Bookmark, U eid %(x)s",
                {"x": user.eid},
            )
            cnx.execute('SET X content "duh?" WHERE X is Comment')
            cnx.execute(
                'DELETE Comment C WHERE C comments Y, Y is Card, Y title "une autre'
                ' news !"'
            )
            # check only one supervision email operation
            sentops = [
                op for op in cnx.pending_operations if isinstance(op, SupervisionMailOp)
            ]
            self.assertEqual(len(sentops), 1)
            # check view content
            op = sentops[0]
            view = sentops[0]._get_view()
            self.assertEqual(view.recipients(), ["test@logilab.fr"])
            self.assertEqual(view.subject(), "[data supervision] changes summary")
            data = view.render(
                changes=cnx.transaction_data.get("pendingchanges")
            ).strip()
            data = re.sub(r"#\d+", "#EID", data)
            data = re.sub(r"/\d+", "/EID", data)
            self.assertMultiLineEqual(
                f"""user admin has made the following change(s):

* added cwuser #EID (toto)
  {BASE_URL}cwuser/toto

* added relation in_group from cwuser #EID to cwgroup #EID

* deleted card #EID (une news !)

* added relation bookmarked_by from bookmark #EID to cwuser #EID

* updated comment #EID (duh?)
  {BASE_URL}comment/EID

* deleted comment #EID (duh?)""",
                data,
            )
            # check prepared email
            op._prepare_email()
            self.assertEqual(len(op.to_send), 1)
            self.assertTrue(op.to_send[0][0])
            self.assertEqual(op.to_send[0][1], ["test@logilab.fr"])
            cnx.commit()
            # some other changes #######
            user.cw_adapt_to("IWorkflowable").fire_transition("deactivate")
            sentops = [
                op for op in cnx.pending_operations if isinstance(op, SupervisionMailOp)
            ]
            self.assertEqual(len(sentops), 1)
            # check view content
            op = sentops[0]
            view = sentops[0]._get_view()
            data = view.render(
                changes=cnx.transaction_data.get("pendingchanges")
            ).strip()
            data = re.sub(r"#\d+", "#EID", data)
            data = re.sub(r"/\d+", "/EID", data)
            self.assertMultiLineEqual(
                f"""user admin has made the following change(s):

* changed state of cwuser #EID (toto)
  from state activated to state deactivated
  {BASE_URL}cwuser/toto""",
                data,
            )

    def test_nonregr1(self):
        with self.admin_access.repo_cnx() as cnx:
            # do some unlogged modification
            cnx.execute(
                "SET X last_login_time NOW WHERE X eid %(x)s", {"x": cnx.user.eid}
            )
            cnx.commit()  # no crash


if __name__ == "__main__":
    unittest_main()



# copyright 2003-2014 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.

from yams.buildobjs import (
    RelationDefinition,
    RelationType,
    EntityType,
    String,
    Datetime,
    Int,
)

from cubicweb import _
from cubicweb.schema import ERQLExpression


class friend(RelationDefinition):
    subject = ("CWUser", "CWGroup")
    object = ("CWUser", "CWGroup")
    symmetric = True


class Folder(EntityType):
    name = String()


class parent(RelationDefinition):
    subject = "Folder"
    object = "Folder"
    composite = "object"
    cardinality = "?*"


class children(RelationDefinition):
    subject = "Folder"
    object = "Folder"
    composite = "subject"


class Email(EntityType):
    """electronic mail"""

    subject = String(fulltextindexed=True)
    date = Datetime(description=_("UTC time on which the mail was sent"))
    messageid = String(required=True, indexed=True)
    headers = String(description=_("raw headers"))


class EmailPart(EntityType):
    """an email attachment"""

    __permissions__ = {
        "read": (
            "managers",
            "users",
            "guests",
        ),  # XXX if E parts X, U has_read_permission E
        "add": (
            "managers",
            ERQLExpression("E parts X, U has_update_permission E"),
        ),
        "delete": ("managers", ERQLExpression("E parts X, U has_update_permission E")),
        "update": (
            "managers",
            "owners",
        ),
    }

    content = String(fulltextindexed=True)
    content_format = String(required=True, maxsize=50)
    ordernum = Int(required=True)


class parts(RelationType):
    subject = "Email"
    object = "EmailPart"
    cardinality = "*1"
    composite = "subject"
    fulltext_container = "subject"


class sender(RelationDefinition):
    subject = "Email"
    object = "EmailAddress"
    cardinality = "?*"
    inlined = True


class recipients(RelationDefinition):
    subject = "Email"
    object = "EmailAddress"



# copyright 2021 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact https://www.logilab.fr/ -- mailto:contact@logilab.fr
#
# This file is part of CubicWeb.
#
# CubicWeb is free software: you can redistribute it and/or modify it under the
# terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# CubicWeb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with CubicWeb.  If not, see <https://www.gnu.org/licenses/>.
"""Unit tests for cw.rqlsuggestions"""

from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.rqlsuggestions import RQLSuggestionsBuilder


class RQLSuggestionsBuilderTC(CubicWebTC):
    def suggestions(self, rql):
        with self.admin_access.cnx() as cnx:
            suggestions_builder = RQLSuggestionsBuilder(cnx)
            return suggestions_builder.build_suggestions(rql)

    def test_no_restrictions_rql(self):
        self.assertListEqual([], self.suggestions(""))
        self.assertListEqual([], self.suggestions("An"))
        self.assertListEqual([], self.suggestions("Any X"))
        self.assertListEqual([], self.suggestions("Any X, Y"))

    def test_invalid_rql(self):
        self.assertListEqual([], self.suggestions("blabla"))
        self.assertListEqual([], self.suggestions("Any X WHERE foo, bar"))

    def test_is_rql(self):
        self.assertListEqual(
            [
                f"Any X WHERE X is {eschema}"
                for eschema in sorted(self.vreg.schema.entities())
                if not eschema.final
            ],
            self.suggestions("Any X WHERE X is"),
        )

        self.assertListEqual(
            ["Any X WHERE X is Personne", "Any X WHERE X is Produit"],
            self.suggestions("Any X WHERE X is P"),
        )

        self.assertListEqual(
            [
                "Any X WHERE X is Personne, Y is Personne",
                "Any X WHERE X is Personne, Y is Produit",
            ],
            self.suggestions("Any X WHERE X is Personne, Y is P"),
        )

    def test_relations_rql(self):
        self.assertListEqual(
            [
                "Any X WHERE X is Personne, X actionnaire A",
                "Any X WHERE X is Personne, X associe A",
                "Any X WHERE X is Personne, X connait A",
                "Any X WHERE X is Personne, X dirige A",
                "Any X WHERE X is Personne, X evaluee A",
                "Any X WHERE X is Personne, X nom A",
                "Any X WHERE X is Personne, X prenom A",
                "Any X WHERE X is Personne, X promo A",
                "Any X WHERE X is Personne, X travaille A",
                "Any X WHERE X is Personne, X type A",
            ],
            self.suggestions("Any X WHERE X is Personne, X "),
        )
        self.assertListEqual(
            [
                "Any X WHERE X is Personne, X travaille A",
                "Any X WHERE X is Personne, X type A",
            ],
            self.suggestions("Any X WHERE X is Personne, X t"),
        )
        # try completion on selected
        self.assertListEqual(
            [
                "Any X WHERE X is Personne, Y is Societe, X travaille Y",
                "Any X WHERE X is Personne, Y is Societe, X type A",
            ],
            self.suggestions("Any X WHERE X is Personne, Y is Societe, X t"),
        )
        # invalid relation should not break
        self.assertListEqual(
            [], self.suggestions("Any X WHERE X is Personne, X asdasd")
        )

    def test_attribute_vocabulary_rql(self):
        self.assertListEqual(
            [
                'Any X WHERE X is Personne, X promo "bon"',
                'Any X WHERE X is Personne, X promo "pasbon"',
            ],
            self.suggestions('Any X WHERE X is Personne, X promo "'),
        )
        self.assertListEqual(
            [
                'Any X WHERE X is Personne, X promo "pasbon"',
            ],
            self.suggestions('Any X WHERE X is Personne, X promo "p'),
        )
        # "bon" should be considered complete, hence no suggestion
        self.assertListEqual(
            [], self.suggestions('Any X WHERE X is Personne, X promo "bon"')
        )
        # no valid vocabulary starts with "po"
        self.assertListEqual(
            [], self.suggestions('Any X WHERE X is Personne, X promo "po')
        )

    def test_attribute_value_rql(self):
        # suggestions should contain any possible value for
        # a given attribute (limited to 10)
        with self.admin_access.cnx() as cnx:
            for i in range(15):
                cnx.create_entity("Personne", nom=f"n{i}", prenom=f"p{i}")
            cnx.commit()
        self.assertListEqual(
            [
                'Any X WHERE X is Personne, X nom "n0"',
                'Any X WHERE X is Personne, X nom "n1"',
                'Any X WHERE X is Personne, X nom "n10"',
                'Any X WHERE X is Personne, X nom "n11"',
                'Any X WHERE X is Personne, X nom "n12"',
                'Any X WHERE X is Personne, X nom "n13"',
                'Any X WHERE X is Personne, X nom "n14"',
                'Any X WHERE X is Personne, X nom "n2"',
                'Any X WHERE X is Personne, X nom "n3"',
                'Any X WHERE X is Personne, X nom "n4"',
                'Any X WHERE X is Personne, X nom "n5"',
                'Any X WHERE X is Personne, X nom "n6"',
                'Any X WHERE X is Personne, X nom "n7"',
                'Any X WHERE X is Personne, X nom "n8"',
                'Any X WHERE X is Personne, X nom "n9"',
            ],
            self.suggestions('Any X WHERE X is Personne, X nom "'),
        )
        self.assertListEqual(
            [
                'Any X WHERE X is Personne, X nom "n1"',
                'Any X WHERE X is Personne, X nom "n10"',
                'Any X WHERE X is Personne, X nom "n11"',
                'Any X WHERE X is Personne, X nom "n12"',
                'Any X WHERE X is Personne, X nom "n13"',
                'Any X WHERE X is Personne, X nom "n14"',
            ],
            self.suggestions('Any X WHERE X is Personne, X nom "n1'),
        )
